@@ -4,10 +4,10 @@ from typing import Callable
 
 from localstack.config import dirs
 from localstack.utils import common
-
-# URL to "cfn-response" module which is required in some CF Lambdas
 from localstack.utils.common import select_attributes, short_uid
 
+# URL to "cfn-response" module which is required in some CF Lambdas. The purpose of cfn-response is to make it easier
+# to write "inline" code for custom resources. TODO: consider copying code into our repo instead of downloading it
 CFN_RESPONSE_MODULE_URL = (
     "https://raw.githubusercontent.com/LukeMizuhashi/cfn-response/master/index.js"
 )
@@ -43,12 +43,12 @@ def param_defaults(param_func, defaults):
 
 
 def remove_none_values(params):
-    """Remove None values recursively in the given object."""
+    """Remove None values and AWS::NoValue placeholders (recursively) in the given object."""
 
     def remove_nones(o, **kwargs):
         if isinstance(o, dict):
             for k, v in dict(o).items():
-                if v is None:
+                if v in [None, PLACEHOLDER_AWS_NO_VALUE]:
                     o.pop(k)
         if isinstance(o, list):
             common.run_safe(o.remove, None)
