@@ -9,12 +9,15 @@ import traceback
 import uuid
 from typing import Dict, List
 
+import botocore.exceptions
 import requests as requests
 from flask import Response as FlaskResponse
 from moto.sns.exceptions import DuplicateSnsEndpointError
+from moto.sns.models import MAXIMUM_MESSAGE_LENGTH
 from requests.models import Response
 
 from localstack.aws.api import RequestContext
+from localstack.aws.api.core import CommonServiceException
 from localstack.aws.api.sns import (
     ActionsList,
     AmazonResourceName,
@@ -116,7 +119,8 @@ SNS_PROTOCOLS = [
 
 # set up logger
 LOG = logging.getLogger(__name__)
-HTTP_SUBSCRIPTION_ATTRIBUTES = ["UnsubscribeURL"]
+
+GCM_URL = "https://fcm.googleapis.com/fcm/send"
 
 
 class SNSBackend(RegionBackend):
@@ -128,6 +132,7 @@ class SNSBackend(RegionBackend):
     sns_tags: Dict[str, List[Dict]]
     # cache of topic ARN to platform endpoint messages (used primarily for testing)
     platform_endpoint_messages: Dict[str, List[Dict]]
+
     # list of sent SMS messages - TODO: expose via internal API
     sms_messages: List[Dict]
 
@@ -148,12 +153,24 @@ def publish_message(
 
     target_arn = req_data.get("TargetArn")
     if target_arn and ":endpoint/" in target_arn:
-        # cache messages published to platform endpoints
         cache = sns_backend.platform_endpoint_messages[target_arn] = (
             sns_backend.platform_endpoint_messages.get(target_arn) or []
         )
-        # TODO: check this in the old provider
         cache.append(req_data)
+        platform_app, endpoint_attributes = get_attributes_for_application_endpoint(target_arn)
+        message_structure = req_data.get("MessageStructure", [None])[0]
+        LOG.debug("Publishing message to Endpoint: %s | Message: %s", target_arn, message)
+
+        start_thread(
+            lambda _: message_to_endpoint(
+                target_arn,
+                message,
+                message_structure,
+                endpoint_attributes,
+                platform_app,
+            )
+        )
+        return message_id
 
     LOG.debug("Publishing message to TopicArn: %s | Message: %s", topic_arn, message)
     start_thread(
@@ -169,9 +186,109 @@ def publish_message(
             message_attributes,
         )
     )
+
     return message_id
 
 
+def get_attributes_for_application_endpoint(target_arn):
+    sns_client = aws_stack.connect_to_service("sns")
+    app_name = target_arn.split("/")[-2]
+
+    endpoint_attributes = None
+    try:
+        endpoint_attributes = sns_client.get_endpoint_attributes(EndpointArn=target_arn)[
+            "Attributes"
+        ]
+    except botocore.exceptions.ClientError:
+        LOG.warning(f"Missing attributes for endpoint: {target_arn}")
+    if not endpoint_attributes:
+        raise CommonServiceException(
+            message="No account found for the given parameters",
+            code="InvalidClientTokenId",
+            status_code=403,
+        )
+
+    platform_apps = sns_client.list_platform_applications()["PlatformApplications"]
+    app = None
+    try:
+        app = [x for x in platform_apps if app_name in x["PlatformApplicationArn"]][0]
+    except IndexError:
+        LOG.warning(f"Missing application: {target_arn}")
+
+    if not app:
+        raise CommonServiceException(
+            message="No account found for the given parameters",
+            code="InvalidClientTokenId",
+            status_code=403,
+        )
+
+    # Validate parameters
+    if "app/GCM/" in app["PlatformApplicationArn"]:
+        validate_gcm_parameters(app, endpoint_attributes)
+
+    return app, endpoint_attributes
+
+
+def message_to_endpoint(target_arn, message, structure, endpoint_attributes, platform_app):
+    if structure == "json":
+        message = json.loads(message)
+
+    platform_name = target_arn.split("/")[-3]
+
+    response = None
+    if platform_name == "GCM":
+        response = send_message_to_GCM(
+            platform_app["Attributes"], endpoint_attributes, message["GCM"]
+        )
+
+    if response is None:
+        LOG.warn("Platform not implemeted yet")
+    elif response.status_code != 200:
+        LOG.warn(
+            f"Platform {platform_name} returned response {response.status_code} with content {response.content}"
+        )
+
+
+def validate_gcm_parameters(platform_app: Dict, endpoint_attributes: Dict):
+    server_key = platform_app["Attributes"].get("PlatformCredential", "")
+    if not server_key:
+        raise InvalidParameterException(
+            "Invalid parameter: Attributes Reason: Invalid value for attribute: PlatformCredential: cannot be empty"
+        )
+    headers = {"Authorization": f"key={server_key}", "Content-type": "application/json"}
+    response = requests.post(
+        GCM_URL,
+        headers=headers,
+        data='{"registration_ids":["ABC"]}',
+    )
+
+    if response.status_code == 401:
+        raise InvalidParameterException(
+            "Invalid parameter: Attributes Reason: Platform credentials are invalid"
+        )
+
+    if not endpoint_attributes.get("Token"):
+        raise InvalidParameterException(
+            "Invalid parameter: Attributes Reason: Invalid value for attribute: Token: cannot be empty"
+        )
+
+
+def send_message_to_GCM(app_attributes, endpoint_attributes, message):
+    server_key = app_attributes.get("PlatformCredential", "")
+    token = endpoint_attributes.get("Token", "")
+    data = json.loads(message)
+
+    data["to"] = token
+    headers = {"Authorization": f"key={server_key}", "Content-type": "application/json"}
+
+    response = requests.post(
+        GCM_URL,
+        headers=headers,
+        data=json.dumps(data),
+    )
+    return response
+
+
 class SnsProvider(SnsApi, ServiceLifecycleHook):
     def add_permission(
         self,
@@ -402,6 +519,7 @@ class SnsProvider(SnsApi, ServiceLifecycleHook):
             for i in v:
                 if i["TopicArn"] == topic_arn:
                     i["PendingConfirmation"] = "false"
+
         return ConfirmSubscriptionResponse(SubscriptionArn=sub_arn)
 
     def untag_resource(
@@ -464,17 +582,15 @@ class SnsProvider(SnsApi, ServiceLifecycleHook):
 
             if current_subscription["Protocol"] in ["http", "https"]:
                 external_url = external_service_url("sns")
-                token = short_uid()
+                subscription_token = short_uid()
                 message_id = long_uid()
-                subscription_url = "%s/?Action=ConfirmSubscription&SubscriptionArn=%s&Token=%s" % (
-                    external_url,
-                    target_subscription_arn,
-                    token,
+                subscription_url = create_subscribe_url(
+                    external_url, current_subscription["TopicArn"], subscription_token
                 )
                 message = {
                     "Type": ["UnsubscribeConfirmation"],
                     "MessageId": [message_id],
-                    "Token": [token],
+                    "Token": [subscription_token],
                     "TopicArn": [current_subscription["TopicArn"]],
                     "Message": [
                         "You have chosen to deactivate subscription %s.\nTo cancel this operation and restore the subscription, visit the SubscribeURL included in this message."
@@ -538,6 +654,9 @@ class SnsProvider(SnsApi, ServiceLifecycleHook):
         if not message or all(not m for m in message):
             raise InvalidParameterException("Empty message")
 
+        if len(message) > MAXIMUM_MESSAGE_LENGTH:
+            raise InvalidParameterException("Message too long")
+
         if topic_arn and ".fifo" in topic_arn and not message_group_id:
             raise InvalidParameterException(
                 "The MessageGroupId parameter is required for FIFO topics",
@@ -586,6 +705,10 @@ class SnsProvider(SnsApi, ServiceLifecycleHook):
             raise InvalidParameterException(
                 f"Invalid parameter: Amazon SNS does not support this protocol string: {protocol}"
             )
+        elif protocol in ["http", "https"] and not endpoint.startswith(f"{protocol}://"):
+            raise InvalidParameterException(
+                "Invalid parameter: Endpoint must match the specified protocol"
+            )
         if ".fifo" in endpoint and ".fifo" not in topic_arn:
             raise InvalidParameterException(
                 "FIFO SQS Queues can not be subscribed to standard SNS topics"
@@ -621,35 +744,28 @@ class SnsProvider(SnsApi, ServiceLifecycleHook):
         if subscription_arn not in sns_backend.subscription_status:
             sns_backend.subscription_status[subscription_arn] = {}
 
-        token = short_uid()
+        subscription_token = short_uid()
         sns_backend.subscription_status[subscription_arn].update(
-            {"TopicArn": topic_arn, "Token": token, "Status": "Not Subscribed"}
+            {"TopicArn": topic_arn, "Token": subscription_token, "Status": "Not Subscribed"}
         )
         # Send out confirmation message for HTTP(S), fix for https://github.com/localstack/localstack/issues/881
         if protocol in ["http", "https"]:
-            token = short_uid()
             external_url = external_service_url("sns")
-            subscription["UnsubscribeURL"] = "%s/?Action=Unsubscribe&SubscriptionArn=%s" % (
-                external_url,
-                subscription_arn,
-            )
+            subscription["UnsubscribeURL"] = create_unsubscribe_url(external_url, subscription_arn)
             confirmation = {
                 "Type": ["SubscriptionConfirmation"],
-                "Token": [token],
+                "Token": [subscription_token],
                 "Message": [
-                    ("You have chosen to subscribe to the topic %s.\n" % topic_arn)
+                    f"You have chosen to subscribe to the topic {topic_arn}.\n"
                     + "To confirm the subscription, visit the SubscribeURL included in this message."
                 ],
-                "SubscribeURL": [
-                    "%s/?Action=ConfirmSubscription&TopicArn=%s&Token=%s"
-                    % (external_url, topic_arn, token)
-                ],
+                "SubscribeURL": [create_subscribe_url(external_url, topic_arn, subscription_token)],
             }
             publish_message(topic_arn, confirmation, {}, subscription_arn, skip_checks=True)
         elif protocol in ["sqs", "lambda"]:
             # Auto-confirm sqs and lambda subscriptions for now
             # TODO: revisit for multi-account
-            self.confirm_subscription(context, topic_arn, token)
+            self.confirm_subscription(context, topic_arn, subscription_token)
         return SubscribeResponse(SubscriptionArn=subscription_arn)
 
     def tag_resource(
@@ -839,7 +955,7 @@ async def message_to_subscriber(
         except Exception as exc:
             LOG.info("Unable to forward SNS message to SQS: %s %s", exc, traceback.format_exc())
             store_delivery_log(subscriber, False, message, message_id)
-            sns_error_to_dead_letter_queue(subscriber["SubscriptionArn"], message_body, str(exc))
+            sns_error_to_dead_letter_queue(subscriber, message_body, str(exc))
             if "NonExistentQueue" in str(exc):
                 LOG.info(
                     'Removing non-existent queue "%s" subscribed to topic "%s"',
@@ -852,10 +968,7 @@ async def message_to_subscriber(
     elif subscriber["Protocol"] == "lambda":
         try:
             external_url = external_service_url("sns")
-            unsubscribe_url = "%s/?Action=Unsubscribe&SubscriptionArn=%s" % (
-                external_url,
-                subscriber["SubscriptionArn"],
-            )
+            unsubscribe_url = create_unsubscribe_url(external_url, subscriber["SubscriptionArn"])
             response = lambda_api.process_sns_notification(
                 subscriber["Endpoint"],
                 topic_arn,
@@ -888,7 +1001,7 @@ async def message_to_subscriber(
             )
             store_delivery_log(subscriber, False, message, message_id)
             message_body = create_sns_message_body(subscriber, req_data, message_id)
-            sns_error_to_dead_letter_queue(subscriber["SubscriptionArn"], message_body, str(exc))
+            sns_error_to_dead_letter_queue(subscriber, message_body, str(exc))
         return
 
     elif subscriber["Protocol"] in ["http", "https"]:
@@ -898,24 +1011,28 @@ async def message_to_subscriber(
         except Exception:
             return
         try:
-            headers = {
+            message_headers = {
                 "Content-Type": "text/plain",
                 # AWS headers according to
                 # https://docs.aws.amazon.com/sns/latest/dg/sns-message-and-json-formats.html#http-header
                 "x-amz-sns-message-type": msg_type,
+                "x-amz-sns-message-id": message_id,
                 "x-amz-sns-topic-arn": subscriber["TopicArn"],
-                "x-amz-sns-subscription-arn": subscriber["SubscriptionArn"],
                 "User-Agent": "Amazon Simple Notification Service Agent",
             }
+            if msg_type != "SubscriptionConfirmation":
+                # while testing, never had those from AWS but the docs above states it should be there
+                message_headers["x-amz-sns-subscription-arn"] = subscriber["SubscriptionArn"]
+
             # When raw message delivery is enabled, x-amz-sns-rawdelivery needs to be set to 'true'
             # indicating that the message has been published without JSON formatting.
             # https://docs.aws.amazon.com/sns/latest/dg/sns-large-payload-raw-message-delivery.html
-            if is_raw_message_delivery(subscriber):
-                headers["x-amz-sns-rawdelivery"] = "true"
+            elif msg_type == "Notification" and is_raw_message_delivery(subscriber):
+                message_headers["x-amz-sns-rawdelivery"] = "true"
 
             response = requests.post(
                 subscriber["Endpoint"],
-                headers=headers,
+                headers=message_headers,
                 data=message_body,
                 verify=False,
             )
@@ -932,7 +1049,9 @@ async def message_to_subscriber(
                 "Received error on sending SNS message, putting to DLQ (if configured): %s", exc
             )
             store_delivery_log(subscriber, False, message, message_id)
-            sns_error_to_dead_letter_queue(subscriber["SubscriptionArn"], message_body, str(exc))
+            # AWS doesnt send to the DLQ if there's an error trying to deliver a UnsubscribeConfirmation msg
+            if msg_type != "UnsubscribeConfirmation":
+                sns_error_to_dead_letter_queue(subscriber, message_body, str(exc))
         return
 
     elif subscriber["Protocol"] == "application":
@@ -948,7 +1067,7 @@ async def message_to_subscriber(
             )
             store_delivery_log(subscriber, False, message, message_id)
             message_body = create_sns_message_body(subscriber, req_data, message_id)
-            sns_error_to_dead_letter_queue(subscriber["SubscriptionArn"], message_body, str(exc))
+            sns_error_to_dead_letter_queue(subscriber, message_body, str(exc))
         return
 
     elif subscriber["Protocol"] in ["email", "email-json"]:
@@ -989,6 +1108,7 @@ def get_subscription_by_arn(sub_arn):
 
 def create_sns_message_body(subscriber, req_data, message_id=None) -> str:
     message = req_data["Message"][0]
+    message_type = req_data.get("Type", ["Notification"])[0]
     protocol = subscriber["Protocol"]
     attributes = get_message_attributes(req_data)
 
@@ -999,18 +1119,15 @@ def create_sns_message_body(subscriber, req_data, message_id=None) -> str:
         except KeyError:
             raise Exception("Unable to find 'default' key in message payload")
 
-    if is_raw_message_delivery(subscriber):
+    if message_type == "Notification" and is_raw_message_delivery(subscriber):
         if attributes:
             message["MessageAttributes"] = attributes
         return message
 
     external_url = external_service_url("sns")
-    unsubscribe_url = (
-        f"{external_url}/?Action=Unsubscribe&SubscriptionArn={subscriber['SubscriptionArn']}"
-    )
 
     data = {
-        "Type": req_data.get("Type", ["Notification"])[0],
+        "Type": message_type,
         "MessageId": message_id,
         "TopicArn": subscriber["TopicArn"],
         "Message": message,
@@ -1020,17 +1137,16 @@ def create_sns_message_body(subscriber, req_data, message_id=None) -> str:
         # Hardcoded
         "Signature": "EXAMPLEpH+..",
         "SigningCertURL": "https://sns.us-east-1.amazonaws.com/SimpleNotificationService-0000000000000000000000.pem",
-        "UnsubscribeURL": unsubscribe_url,
     }
 
+    if message_type == "Notification":
+        unsubscribe_url = create_unsubscribe_url(external_url, subscriber["SubscriptionArn"])
+        data["UnsubscribeURL"] = unsubscribe_url
+
     for key in ["Subject", "SubscribeURL", "Token"]:
         if req_data.get(key) and req_data[key][0]:
             data[key] = req_data[key][0]
 
-    for key in HTTP_SUBSCRIPTION_ATTRIBUTES:
-        if key in subscriber:
-            data[key] = subscriber[key]
-
     if attributes:
         data["MessageAttributes"] = attributes
 
@@ -1091,6 +1207,14 @@ def prepare_message_attributes(message_attributes):
     return attributes
 
 
+def create_subscribe_url(external_url, topic_arn, subscription_token):
+    return f"{external_url}/?Action=ConfirmSubscription&TopicArn={topic_arn}&Token={subscription_token}"
+
+
+def create_unsubscribe_url(external_url, subscription_arn):
+    return f"{external_url}/?Action=Unsubscribe&SubscriptionArn={subscription_arn}"
+
+
 def is_number(x):
     try:
         float(x)
